/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
/* Spatial stereo test. */



/* The following description is used by demos/make-demo-list.pl
 * to compile a list of all tests. */
/* DESCRIPTION
 * Spatial stereo test.
 *
 * Function: demo_spatial_stereo_init
 */


#include <gtk/gtk.h>
#include <gdam/gdambuffer.h>
#include "../gdamchannel.h"

typedef enum _DemoDragMode {
	DRAGGING_NONE,
	DRAGGING_CENTER,
	DRAGGING_LEFT,
	DRAGGING_RIGHT
} DemoDragMode;

typedef struct _DemoInfo {
	GdamChannel* 		channel;
	GtkWindow* 		top_level;
	GtkLabel* 		filename;
	GtkLabel* 		state;
	GtkFileSelection* 	file_selection;
	guint 			mixer_id;
	guint 			spatial_id;
	guint 			subscription_id;
	gboolean 		making_spatial_request;

	gdouble			center_pos[3];
	gdouble			right_offset[3];
	int			drag_offset_x, drag_offset_y;

	int			center_radius;
	int			channel_radius;
	GdkGC*			background;
	GdkGC*			circle_gc;

	DemoDragMode		drag_mode;

	GdkPixmap*		backing_pixmap;
	/* Height and width of the drawing area. */
	int			draw_width, draw_height;
	/* The drawing area for the samples... */
	GtkWidget*		drawing_area;

	/* The boundaries of the world. */
	gdouble			world_left, world_top;
	gdouble			world_width, world_height;
} DemoInfo;

static DemoInfo* new_blank_demo_info()
{
	DemoInfo* info = g_new0(DemoInfo, 1);
	info->center_pos[0] = 0.0;
	info->center_pos[1] = 2.0;
	info->center_pos[2] = 0.0;
	info->right_offset[0] = 0.2;
	info->right_offset[1] = 0.0;
	info->right_offset[2] = 0.0;
	info->world_left = info->world_top = -10.0;
	info->world_width = info->world_height = +20.0;
	info->channel_radius = 4;
	info->center_radius = 8;
	return info;
}


static void reset_spatial_parameters(DemoInfo* info);
static void get_left_channel_knob_position (DemoInfo* info, int* x, int* y);
static void get_right_channel_knob_position(DemoInfo* info, int* x, int* y);
static void get_center_knob_position(DemoInfo* info, int* x, int* y);


/*=====================================================================
 * Functions to render the draw area. 
 */
static void render_foreground(DemoInfo* info)
{
	GdkPixmap* pixmap = info->backing_pixmap;
	int x, y, radius;
	radius = info->center_radius;

	get_left_channel_knob_position(info, &x, &y);
	radius = info->channel_radius;
	g_message("draw_arc: %d,%d (radius %d)", x,y,radius);
	gdk_draw_arc(pixmap,
	             info->circle_gc,
		     TRUE,
		     x - radius + 1,
		     y - radius + 1,
		     radius * 2 - 1,
		     radius * 2 - 1,
		     0,
		     360*64);

	get_right_channel_knob_position(info, &x, &y);
	radius = info->channel_radius;
	g_message("draw_arc: %d,%d (radius %d)", x,y,radius);
	gdk_draw_arc(pixmap,
	             info->circle_gc,
		     TRUE,
		     x - radius + 1,
		     y - radius + 1,
		     radius * 2 - 1,
		     radius * 2 - 1,
		     0,
		     360*64);

	get_center_knob_position(info, &x, &y);
	radius = info->center_radius;
	g_message("draw_arc: %d,%d (radius %d)", x,y,radius);
	gdk_draw_arc(pixmap,
	             info->circle_gc,
		     TRUE,
		     x - radius + 1,
		     y - radius + 1,
		     radius * 2 - 1,
		     radius * 2 - 1,
		     0,
		     360*64);
}
static gint
configure_event(GtkWidget* widget,
                GdkEventConfigure* configure_event,
		DemoInfo* info)
{
	if (info->backing_pixmap)
		gdk_pixmap_unref(info->backing_pixmap);
	info->backing_pixmap = gdk_pixmap_new(widget->window,
	                              widget->allocation.width,
	                              widget->allocation.height,
				      -1);
	info->draw_width = widget->allocation.width;
	info->draw_height = widget->allocation.height;
	return TRUE;
}

static void alloc_colors(DemoInfo* info)
{
	int state = GTK_WIDGET_STATE(info->drawing_area);
	/** HACK!!! */
	info->circle_gc = info->drawing_area->style->bg_gc[state];
	info->background = info->drawing_area->style->fg_gc[state];
}

static void copy_backing_area(DemoInfo* info, int x, int y, int w, int h)
{
        gdk_draw_pixmap(info->drawing_area->window,
                        info->drawing_area->style
				->fg_gc[GTK_WIDGET_STATE(info->drawing_area)],
                        info->backing_pixmap,
                        x, y, x, y, w, h);
}

static void redraw_area(DemoInfo* info, int x, int y, int w, int h)
{
	if (!info->circle_gc)
		alloc_colors(info);
	g_message("redraw_area: %d,%d x %d,%d", x,y,w,h);
	gdk_draw_rectangle(info->backing_pixmap,
	                   info->background,
		           TRUE,
		           x,y,w,h);
	render_foreground(info);
	copy_backing_area(info, x, y, w, h);
}

static gint
expose_event(GtkDrawingArea* area,
             GdkEventExpose* expose_event,
             DemoInfo* info)
{
	redraw_area(info, 
		    expose_event->area.x, expose_event->area.y,
                    expose_event->area.width, expose_event->area.height);
	return 0;
}

static int world_to_screen_x(DemoInfo* info, gdouble x)
{
	return (x - info->world_left) * info->draw_width / info->world_width;
}

static int world_to_screen_y(DemoInfo* info, gdouble y)
{
	return (y - info->world_top) * info->draw_height / info->world_height;
}

static void screen_to_world(DemoInfo* info,
                            int x, int y,
			    gdouble* position_out)
{
	gdouble world_x, world_y;

	world_x = x;
	world_x /= info->draw_width;
	world_x *= info->world_width;
	world_x += info->world_left;
	position_out[0] = world_x;

	world_y = y;
	world_y /= info->draw_height;
	world_y *= info->world_height;
	world_y += info->world_top;
	position_out[1] = world_y;

	position_out[2] = 0.0;
}
static void get_left_channel_knob_position (DemoInfo*    info,
                                            int*         x,
					    int*         y)
{
	*x = world_to_screen_x(info, info->center_pos[0]-info->right_offset[0]);
	*y = world_to_screen_y(info, info->center_pos[1]-info->right_offset[1]);
}

static void get_right_channel_knob_position(DemoInfo*    info,
                                            int*         x,
					    int*         y)
{
	*x = world_to_screen_x(info, info->center_pos[0]+info->right_offset[0]);
	*y = world_to_screen_y(info, info->center_pos[1]+info->right_offset[1]);
}

static void get_center_knob_position(DemoInfo*    info,
                                     int*         x,
				     int*         y)
{
	*x = world_to_screen_x(info, info->center_pos[0]);
	*y = world_to_screen_y(info, info->center_pos[1]);
}

static void compute_rendered_area(DemoInfo* info,
                                  GdkRectangle* rectangle)
{
	int x, y;
	GdkRectangle total;
	GdkRectangle tmp;
	GdkRectangle last;

	
	get_center_knob_position(info, &x, &y);
	last.x = x - info->center_radius;
	last.y = y - info->center_radius;
	last.width = last.height = 2 * info->center_radius;

	get_left_channel_knob_position(info, &x, &y);
	tmp.x = x - info->channel_radius;
	tmp.y = y - info->channel_radius;
	tmp.width = tmp.height = 2 * info->channel_radius;
	gdk_rectangle_union(&tmp, &last, &total);
	last = total;

	get_right_channel_knob_position(info, &x, &y);
	tmp.x = x - info->channel_radius;
	tmp.y = y - info->channel_radius;
	tmp.width = tmp.height = 2 * info->channel_radius;
	gdk_rectangle_union(&tmp, &last, &total);

	*rectangle = total;
}

static inline gboolean
is_in_radius(int x1, int y1, int x2, int y2, int maxdist)
{
	int dist_squared = (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2);
	int max_dist_squared = maxdist * maxdist;
	g_message("is_in_radius: (%d,%d) - (%d,%d) [maxdist %d]",
		x1,y1,x2,y2,maxdist);
	return dist_squared <= max_dist_squared;
}
/*=====================================================================
 * Dragging the knobs.
 */
static int mouse_down_event(GtkWidget*        drawing_area,
                            GdkEventButton*   event,
			    DemoInfo*         info)
{
	int obj_x, obj_y;
	int mouse_x = event->x, mouse_y = event->y;
	if (info->drag_mode != DRAGGING_NONE)
		return TRUE;

/*
	if (event->button != GDK_BUTTON1) {
		g_warning("not button1");
		return TRUE;
	}
*/

	get_center_knob_position(info, &obj_x, &obj_y);
	if (is_in_radius(mouse_x, mouse_y,
	                 obj_x, obj_y, info->channel_radius)) {
		info->drag_mode = DRAGGING_CENTER;
		goto found_object;
	}

	get_left_channel_knob_position(info, &obj_x, &obj_y);
	if (is_in_radius(mouse_x, mouse_y,
	                 obj_x, obj_y, info->channel_radius)) {
		info->drag_mode = DRAGGING_LEFT;
		goto found_object;
	}

	get_right_channel_knob_position(info, &obj_x, &obj_y);
	if (is_in_radius(mouse_x, mouse_y,
	                 obj_x, obj_y, info->channel_radius)) {
		info->drag_mode = DRAGGING_RIGHT;
		goto found_object;
	}

	return TRUE;

found_object:
	info->drag_offset_x = obj_x - mouse_x;
	info->drag_offset_y = obj_y - mouse_y;
	return TRUE;
}

static int mouse_up_event(GtkWidget*        drawing_area,
                          GdkEventButton*   event,
			  DemoInfo*         info)
{
	info->drag_mode = DRAGGING_NONE;
	return FALSE;
}

static int mouse_move_event(GtkWidget*        drawing_area,
                            GdkEventMotion*   event,
			    DemoInfo*         info)
{
	GdkRectangle initial_area, final_area, total_area;
	gdouble world_position[3];
	guint state;
	int x, y;

	if (event->is_hint)
		gdk_window_get_pointer(event->window, &x, &y, &state);
	else {
		x = event->x;
		y = event->y;
		state = event->state;
	}

	/* We aren't interested in this event afterall. */
	if ((state & GDK_BUTTON1_MASK) != GDK_BUTTON1_MASK)
		return TRUE;

	/* They are holding down the button, but not dragging anything. */
	if (info->drag_mode == DRAGGING_NONE)
		return TRUE;

	/* Get the area we need to clear: */
	compute_rendered_area(info, &initial_area);

	/* Something gets a new position */
	screen_to_world(info,
	                x + info->drag_offset_x,
	                y + info->drag_offset_y,
		        world_position);
	
	switch (info->drag_mode) {
		case DRAGGING_LEFT:
			info->right_offset[0] = info->center_pos[0]
			                      - world_position[0];
			info->right_offset[1] = info->center_pos[1]
			                      - world_position[1];
			break;
		case DRAGGING_RIGHT:
			info->right_offset[0] = world_position[0]
                                              - info->center_pos[0];
			info->right_offset[1] = world_position[1]
 					      - info->center_pos[1];
			break;
		case DRAGGING_CENTER:
			info->center_pos[0] = world_position[0];
			info->center_pos[1] = world_position[1];
			break;
		default: g_assert_not_reached();
	}


	/* Get the area we need to draw: */
	compute_rendered_area(info, &final_area);
	gdk_rectangle_union(&initial_area, &final_area, &total_area);

	reset_spatial_parameters(info);
	redraw_area(info,
	            total_area.x,
	            total_area.y,
	            total_area.width,
	            total_area.height);

	return TRUE;
}

/*=====================================================================
 * Choosing the file.
 */

static void chose_file(GtkButton* button, DemoInfo* info)
{
	char* filename;
	g_return_if_fail(info->file_selection != NULL);
	filename = gtk_file_selection_get_filename(info->file_selection);
	gtk_label_set_text(info->filename, filename);
}

static void filesel_button_clicked(GtkButton* button, DemoInfo* info)
{
	if (info->file_selection) {
		GtkObject* to_kill = GTK_OBJECT(info->file_selection);
		info->file_selection = NULL;
		gtk_object_destroy(to_kill);
	} else {
		GtkWidget* filesel = gtk_file_selection_new("pick your mp3");
		GtkFileSelection* file_select = (GtkFileSelection*)filesel;
		info->file_selection = file_select;
		gtk_signal_connect(GTK_OBJECT(file_select->ok_button),
		                   "clicked",
		                   (GtkSignalFunc) chose_file,
				   info);
		/*gtk_signal_connect_object(
		                   GTK_OBJECT(file_select->cancel_button),
		                   "clicked",
				   (GtkSignalFunc) gtk_widget_destroy,
				   GTK_OBJECT(file_select));*/
		gtk_widget_show(GTK_WIDGET(file_select));
	}
}

static void indicate(DemoInfo* info, const char* state)
{
	gtk_label_set_text(info->state, state);
}

static void stopped_callback(gpointer user_data)
{
	DemoInfo* info = (DemoInfo*)user_data;
	if (info->spatial_id == 0 || info->making_spatial_request)
		return;
	info->spatial_id = 0;
	indicate(info, "ended");
}
	

/*=====================================================================
 * Creating and stopping the sources.
 */
static void play_callback(GtkButton* button, DemoInfo* info)
{
	gchar* filename;
	int mp3_id;
	gdouble left_ear_pos[3];
	gdouble right_ear_pos[3];

	left_ear_pos[0] = -0.1;
	left_ear_pos[1] = left_ear_pos[2] = 0.0;
	right_ear_pos[0] = +0.1;
	right_ear_pos[1] = right_ear_pos[2] = 0.0;

	if (info->spatial_id != 0 || info->making_spatial_request)
		return;
	if (info->subscription_id) {
		gdam_channel_unsubscribe(info->channel, info->subscription_id);
		info->subscription_id = 0;
	}

	info->making_spatial_request = TRUE;
	gtk_label_get(info->filename, &filename);
	gdam_debug("create spatial??? %s", filename);
	mp3_id = gdam_channel_create_source_generic(
				info->channel,
				"GdamSeekableMp3",
				2,
				"filename",
				GDAM_ARG_TYPE_STRING,
				filename,
				"length",
				GDAM_ARG_TYPE_INT64,
				410000);
	info->spatial_id = gdam_channel_create_source_generic(
				info->channel,
				"GdamSpatialStereo",
				4,
				"filtered",
				GDAM_ARG_TYPE_INT,
				mp3_id,
				"left_ear",
				GDAM_ARG_TYPE_DOUBLE_ARRAY,
				3,
				left_ear_pos,
				"right_ear",
				GDAM_ARG_TYPE_DOUBLE_ARRAY,
				3,
				right_ear_pos,
				"unit_gain_distance",
				GDAM_ARG_TYPE_DOUBLE,
				1.0);
	g_assert(info->spatial_id != 0);
	gdam_debug("gdam_channel_mix_add being called...");
	if (!gdam_channel_mix_add(info->channel,
	                          info->mixer_id,
				  info->spatial_id,
				  -1))
		g_warning("mix_add failed");
	else
		indicate(info, "playing");
	info->subscription_id = gdam_channel_subscribe_source_ended(
						info->channel, 
						info->spatial_id,
						stopped_callback,
						info);
	info->making_spatial_request = FALSE;
}

static void stop_callback(GtkButton* button, DemoInfo* info)
{
	if (info->spatial_id == 0 || info->making_spatial_request)
		return;
	if (!gdam_channel_detach_source(info->channel, info->spatial_id))
		g_warning("detach source failed");
	else
		indicate(info, "stopped");
	info->spatial_id = 0;
}

static void do_nothing(gpointer unused)
{
}

static void reset_spatial_parameters(DemoInfo* info)
{
	GdamRequest* request;
	gdouble left_channel_pos[3], right_channel_pos[3];
	GdamArg* args;
	int i;
	if (info->spatial_id == 0)
		return;
	for (i = 0; i < 3; i++) {
		gdouble offset = info->right_offset[i];
		left_channel_pos[i] = info->center_pos[i] - offset;
		right_channel_pos[i] = info->center_pos[i] + offset;
	}

	args = gdam_arg_new_from_valist(2,
					"left_position",
					GDAM_ARG_TYPE_DOUBLE_ARRAY,
					3,
					left_channel_pos,
					"right_position",
					GDAM_ARG_TYPE_DOUBLE_ARRAY,
					3,
					right_channel_pos);
	request = gdam_request_set_args(info->spatial_id,
	                                2,
					args,
					do_nothing,
					do_nothing,
					NULL);

	if (!gdam_channel_process_request(info->channel, request))
		g_warning("process request failed");
	
	gdam_arg_free_array(2, args);
}


/*=====================================================================
 * Assemblying the widgets.
 */

/* Helper + macro to cut down repetition in demo1_init(). */
static GtkWidget* _generic_button(const char* label, GtkSignalFunc callback,
                                 gpointer user_data)
{
	GtkWidget* button = gtk_button_new_with_label(label);
	gtk_signal_connect(GTK_OBJECT(button), "clicked", callback, user_data);
	gtk_widget_show(GTK_WIDGET(button));
	return button;
}
#define generic_button(label, callback)				\
		_generic_button(label, GTK_SIGNAL_FUNC(callback), info)

void demo_spatial_stereo_init(GdamChannel*           channel,
                guint                  mixer_id)
{
	DemoInfo* info = new_blank_demo_info();
	GtkWidget* hbox;
	GtkWidget* vbox;
	GtkWidget* drawing_area;
	vbox = gtk_vbox_new(FALSE, 0);
	hbox = gtk_hbox_new(FALSE, 0);

	info->channel = channel;
	info->top_level = (GtkWindow*) gtk_window_new(GTK_WINDOW_TOPLEVEL);
	info->filename = (GtkLabel*) gtk_label_new("???");
	info->state = (GtkLabel*) gtk_label_new("not started");
	info->file_selection = NULL;
	info->making_spatial_request = FALSE;
	info->mixer_id = mixer_id;
	info->spatial_id = 0;
	info->subscription_id = 0;

	gtk_widget_show(GTK_WIDGET(info->filename));
	gtk_box_pack_start_defaults(GTK_BOX(hbox), GTK_WIDGET(info->filename));
	gtk_box_pack_start_defaults(GTK_BOX(hbox),
	                  generic_button("File...", filesel_button_clicked));
	gtk_widget_show(hbox);
	gtk_box_pack_start_defaults(GTK_BOX(hbox), GTK_WIDGET(info->state));
	gtk_widget_show(GTK_WIDGET(info->state));

	gtk_box_pack_start_defaults(GTK_BOX(vbox), GTK_WIDGET(hbox));
	info->drawing_area = drawing_area = gtk_drawing_area_new();
	gtk_drawing_area_size(GTK_DRAWING_AREA(drawing_area), 800, 300);
	gtk_signal_connect(GTK_OBJECT(drawing_area),
	                   "button_press_event",
			   GTK_SIGNAL_FUNC(mouse_down_event),
			   info);
	gtk_signal_connect(GTK_OBJECT(drawing_area),
	                   "button_release_event",
			   GTK_SIGNAL_FUNC(mouse_up_event),
			   info);
	gtk_signal_connect(GTK_OBJECT(drawing_area),
	                   "motion_notify_event",
			   GTK_SIGNAL_FUNC(mouse_move_event),
			   info);
	gtk_signal_connect(GTK_OBJECT(drawing_area),
	                   "expose_event",
			   GTK_SIGNAL_FUNC(expose_event),
			   info);
	gtk_signal_connect(GTK_OBJECT(drawing_area),
	                   "configure_event",
			   GTK_SIGNAL_FUNC(configure_event),
			   info);
	gtk_widget_show(info->drawing_area);
	gtk_box_pack_start_defaults(GTK_BOX(vbox), info->drawing_area);
	gtk_widget_add_events(drawing_area, GDK_EXPOSURE_MASK
                                          | GDK_LEAVE_NOTIFY_MASK
                                          | GDK_BUTTON_PRESS_MASK
                                          | GDK_BUTTON_RELEASE_MASK
                                          | GDK_POINTER_MOTION_MASK
                                          | GDK_POINTER_MOTION_HINT_MASK);


	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start_defaults(GTK_BOX(hbox),
	                        generic_button("Play", play_callback));
	gtk_box_pack_start_defaults(GTK_BOX(hbox),
	                        generic_button("Stop", stop_callback));
	gtk_widget_show(hbox);
	gtk_box_pack_start_defaults(GTK_BOX(vbox), GTK_WIDGET(hbox));
	gtk_widget_show(vbox);
	gtk_container_add(GTK_CONTAINER(info->top_level), GTK_WIDGET(vbox));
	gtk_widget_show(GTK_WIDGET(info->top_level));
}
